﻿using System;
using System.Diagnostics;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Threading;
using SongVocalSectionAnalyser.Common.Logging.Model;

namespace SongVocalSectionAnalyser.Common.Logging
{
    public abstract class LoggerBase : ILogger
    {
        private void LogInternal(string summary, object details, [CallerMemberName] string level = null)
        {
            var currentThread = Thread.CurrentThread;

            var stackFrames = new StackTrace(true).GetFrames() ?? new StackFrame[0];
            var stackFrameList = stackFrames.ToList();

            var idx = stackFrameList.FindLastIndex(p =>
            {
                var type = p.GetMethod().DeclaringType;
                return typeof(Logger).IsAssignableFrom(type)
                    || typeof(ILogger).IsAssignableFrom(type);
            });

            if (-1 != idx)
            {
                stackFrameList.RemoveRange(0, idx + 1);
                stackFrames = stackFrameList.ToArray();
            }

            var entry = new LogEntry
            {
                Id = Guid.NewGuid(),
                Level = level?.ToUpper() ?? "UNKNOWN",
                Time = DateTime.Now,
                Thread = new LogThreadInfo
                {
                    Id = currentThread.ManagedThreadId,
                    Name = currentThread.Name ?? "(UNNAMED)"
                },
                Summary = summary,
                Details = details,
                Stacktrace = stackFrames.Select(p =>
                {
                    var method = p.GetMethod();
                    var type = method.DeclaringType;
                    return new LogStacktraceFrame
                    {
                        Assembly = type?.AssemblyQualifiedName ?? "<assembly unknown>",
                        Method = $"{type?.FullName ?? "<class unknown>"}.{method.Name}",
                        Line = new LogStacktraceFrameLineInfo
                        {
                            FileName = p.GetFileName() ?? "<filename unknown>",
                            Line = p.GetFileLineNumber(),
                            Column = p.GetFileColumnNumber()
                        },
                        Arguments = method.GetParameters().Select(q => new LogStacktraceFrameArgument
                        {
                            Type = q.ParameterType.FullName,
                            Name = q.Name
                        }).ToArray()
                    };
                }).ToArray()
            };

            WriteLog(entry);
        }

        protected abstract void WriteLog(LogEntry entry);

        public void Debug(string summary, object details = null) => LogInternal(summary, details);

        public void Trace(string summary, object details = null) => LogInternal(summary, details);

        public void Info(string summary, object details = null) => LogInternal(summary, details);

        public void Warn(string summary, object details = null) => LogInternal(summary, details);

        public void Error(string summary, object details = null) => LogInternal(summary, details);

        public void Fatal(string summary, object details = null) => LogInternal(summary, details);
    }
}
